import os
import tomllib
from collections.abc import Generator, Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from unistrant.error import UnistrantError

default_logging_settings = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": r"%(message)s",
        },
    },
    "handlers": {
        "console": {
            "formatter": "standard",
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "": {
            "handlers": ["console"],
            "level": "INFO",
        },
    },
}


@dataclass(frozen=True)
class Settings(Mapping):
    settings: Mapping[str, Any]
    logging: Mapping[str, Any]

    def __contains__(self, key: Any) -> bool:
        return key in self.settings

    def __getitem__(self, key: Any) -> Any:
        return self.settings[key]

    def __iter__(self) -> Generator:
        yield from self.settings

    def __len__(self) -> int:
        return len(self.settings)

    def required(self, key: Any) -> Any:
        try:
            return self.settings[key]
        except KeyError:
            raise UnistrantError(f"Required setting {key} missing")


def find_settings(settings_path: Path | None) -> Path | None:
    if settings_path:
        return settings_path

    if home := os.getenv("HOME"):
        home_settings_path = (Path(home) / ".config/sams/unistrant/unistrant.toml").expanduser()
        if home_settings_path.exists():
            return home_settings_path

    system_settings_path = Path("/etc/sams/unistrant/unistrant.toml")
    if system_settings_path.exists():
        return system_settings_path

    return None


def load_settings(settings_path: Path | None) -> Settings:
    if settings_path is not None:
        try:
            with settings_path.open(mode="rb") as settings_file:
                settings = tomllib.load(settings_file)
        except OSError as e:
            raise UnistrantError(f"Unable to read settings file: {str(e)}")
        except tomllib.TOMLDecodeError as e:
            raise UnistrantError(f"Unable to decode settings file: {str(e)}")
    else:
        settings = {}

    def flatten(s: dict, base: str = "") -> Any:
        for key, value in s.items():
            full_key = f"{base}.{key}" if base else key
            if isinstance(value, dict):
                yield from flatten(value, full_key)
            else:
                yield full_key, value

    return Settings(settings=dict(flatten(settings)), logging=settings.get("logging", default_logging_settings))



import datetime
import itertools
import logging
import shutil
import uuid
from abc import ABC, abstractmethod
from collections.abc import Generator

from unistrant.http import CertificateAuthentication, HttpProtocol
from unistrant.options import Options
from unistrant.record import Record, RecordDocument, RecordFile, RecordType
from unistrant.sams import SamsClient

logger = logging.getLogger(__name__)


class BaseCommand(ABC):
    def __init__(self, options: Options):
        super().__init__()
        self.options = options
        self.error = False

        authentication = CertificateAuthentication(options.sams_certificate, options.sams_key)
        protocol = HttpProtocol(authentication)
        self.sams = SamsClient(options.sams_url, protocol)

    @abstractmethod
    def run(self) -> None:
        pass

    def fail(self) -> None:
        if not self.error:
            self.error = True


class RegisterCommand(BaseCommand):
    def __init__(self, options: Options):
        super().__init__(options)

        self.record_files = {RecordFile(path) for path in self.options.records_directory.iterdir() if path.is_file()}

    @property
    def records(self) -> Generator[Record, None, None]:
        for file in self.record_files:
            yield from file.records

    @property
    def record_documents(self) -> Generator[RecordDocument, None, None]:
        def record_type_key(record: Record) -> RecordType:
            return record.record_type

        sorted_records = sorted(self.records, key=record_type_key)
        for record_type, iterator in itertools.groupby(sorted_records, key=record_type_key):
            for records in itertools.batched(iterator, 100):
                yield RecordDocument(record_type, records)

    def run(self) -> None:
        documents = set(self.record_documents)
        failed_documents = set()

        total_records = sum([len(document.records) for document in documents])
        total_documents = len(documents)
        logger.info(f"Registering a total of {total_records} records in {total_documents} documents")

        for index, document in enumerate(documents, start=1):
            try:
                logger.info(f"({index}/{total_documents}) Registering {len(document.records)} records of type {document.record_type}")
                self.sams.upload_record_document(document)
            except Exception as e:
                logger.error(f"Error uploading document: {str(e)}")
                failed_documents.add(document)
                self.fail()

        for file in self.record_files:
            self.archive(file)

        for document in failed_documents:
            self.save_for_later(document)

        total_documents_failed = len(failed_documents)
        total_documents_success = total_documents - total_documents_failed
        if total_documents_success > 0:
            logger.info(f"{total_documents_success} documents successfully registered")
        if total_documents_failed > 0:
            logger.error(f"{total_documents_failed} documents unsuccessfully registered")

    def save_for_later(self, document: RecordDocument) -> None:
        destination = self.options.records_directory / f"error-{str(uuid.uuid4())}.xml"
        logger.debug(f"Saving {len(document.records)} records to {destination} for later attempt")
        with destination.open("wb") as f:
            f.write(document.bytes)

    def archive(self, file: RecordFile) -> None:
        if (destination := self.options.archive_directory / file.name).exists():
            timestamp = datetime.datetime.now(datetime.UTC).astimezone().isoformat()
            destination = destination.with_name(f"{file.name}-{timestamp}")
        logger.debug(f"Archiving {file.name} to {destination}")
        shutil.move(file.path, destination)



from abc import ABC, abstractmethod
from collections.abc import Mapping
from pathlib import Path
from typing import Any

import requests


class Authentication(ABC):
    @abstractmethod
    def get_requests_kwargs(self) -> Mapping[str, Any]:
        pass


class CertificateAuthentication(Authentication):
    def __init__(self, certificate: Path, key: Path):
        super().__init__()
        self.certificate = certificate
        self.key = key

    def get_requests_kwargs(self) -> Mapping[str, Any]:
        return {
            "cert": (self.certificate, self.key),
        }


class HttpProtocol:
    def __init__(self, authentication: Authentication | None = None):
        requests_kwargs: dict[str, Any] = {
            "headers": {
                "User-Agent": "Unistrant",
            },
            "timeout": (10, 10),
        }
        if authentication:
            requests_kwargs |= authentication.get_requests_kwargs()
        self.requests_kwargs: Mapping[str, Any] = requests_kwargs

    def get(self, url: str) -> bytes:
        response = requests.get(url, **self.requests_kwargs)
        response.raise_for_status()
        return response.content

    def post(self, url: str, payload: bytes) -> None:
        response = requests.post(url, data=payload, **self.requests_kwargs)
        response.raise_for_status()



import enum
from dataclasses import dataclass
from pathlib import Path

from unistrant.settings import Settings


class CommandName(enum.StrEnum):
    Register = "register"


@dataclass(frozen=True, kw_only=True)
class Options:
    archive_directory_name: str
    command: CommandName
    data_directory: Path
    log_level: str
    records_directory_name: str
    sams_certificate: Path
    sams_key: Path
    sams_url: str
    settings_path: Path | None
    settings: Settings

    @property
    def archive_directory(self) -> Path:
        return self.data_directory / self.archive_directory_name

    @property
    def records_directory(self) -> Path:
        return self.data_directory / self.records_directory_name



from unistrant.app import main

if __name__ == "__main__":
    # Support invoking as a module. The script launcher invokes main
    # directly, so no special initialization should be performed here.
    main()



import argparse
import locale
import logging
import logging.config
import os
import sys
from collections.abc import Mapping
from pathlib import Path
from typing import Any

from unistrant.command import BaseCommand, RegisterCommand
from unistrant.error import UnistrantError
from unistrant.options import CommandName, Options
from unistrant.settings import find_settings, load_settings

logger = logging.getLogger(__name__)


def configure_logging(logging_config: Mapping[str, Any], log_level: str) -> None:
    if logging_config:
        logging.config.dictConfig(dict(logging_config))
    else:
        logging.basicConfig(format="%(message)s")
    if log_level:
        logging.getLogger().setLevel(log_level)


def check_directory(directory: Path) -> None:
    if not directory.exists():
        raise UnistrantError(f"{directory} does not exist")
    if not directory.is_dir():
        raise UnistrantError(f"{directory} is not a directory")


def command_line_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="unistrant", description="This program is a client for the SAMS accounting service.")
    log_levels = ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL")
    parser.add_argument("--archive-directory-name", help="name of archive subdirectory")
    parser.add_argument("--data-directory", type=Path, help="path to data directory")
    parser.add_argument("--log-level", choices=log_levels, type=str.upper, help="override log level")
    parser.add_argument("--records-directory-name", help="name of records subdirectory")
    parser.add_argument("--sams-certificate", type=Path, help="path to client certificate")
    parser.add_argument("--sams-key", type=Path, help="path to client certificate key")
    parser.add_argument("--sams-url", help="accounting service URL")
    parser.add_argument("--settings", type=Path, help="path to settings file")
    parser.add_argument(
        "--use-environment",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="use environment variables (default --use-environment)",
    )

    subparsers = parser.add_subparsers(dest="command", metavar="command", required=True)
    subparsers.add_parser(
        CommandName.Register,
        help="submit records to accounting service",
        description="Use this command to submit record files to the accounting service.",
    )

    return parser


def initialize() -> Options:
    locale.setlocale(locale.LC_ALL, "")

    parser = command_line_parser()
    command_line = parser.parse_args()

    def environment(*key: str) -> str | None:
        if not command_line.use_environment:
            return None

        key_environment = f"UNISTRANT_{'_'.join(key)}".upper()
        return os.getenv(key_environment)

    if settings_environment := environment("settings"):
        settings_override_path = Path(settings_environment)
    else:
        settings_override_path = command_line.settings

    settings_path = find_settings(settings_override_path)
    settings = load_settings(settings_path)

    def option(*key: str, default: Any | None = None, optional: bool = False) -> Any:
        if value := environment(*key):
            return value

        key_command_line = "_".join(key)
        try:
            if value := getattr(command_line, key_command_line):
                return value
        except AttributeError:
            pass

        key_setting = ".".join(key)
        try:
            if value := settings[key_setting]:
                return value
        except KeyError:
            pass

        if optional or default is not None:
            return default

        raise UnistrantError(f"Missing required setting {key_setting}")

    options = Options(
        archive_directory_name=option("archive_directory_name", default="archive"),
        command=CommandName(command_line.command),
        data_directory=option("data_directory"),
        log_level=option("log_level", optional=True),
        records_directory_name=option("records_directory_name", default="records"),
        sams_certificate=option("sams", "certificate"),
        sams_key=option("sams", "key"),
        sams_url=option("sams", "url", default="https://accounting.naiss.se:6143/sgas"),
        settings_path=settings_path,
        settings=settings,
    )

    return options


def _main() -> None:
    options = initialize()

    configure_logging(options.settings.logging, options.log_level)

    check_directory(options.archive_directory)
    check_directory(options.records_directory)

    command: BaseCommand
    match options.command:
        case CommandName.Register:
            command = RegisterCommand(options)

    try:
        command.run()
    finally:
        if command.error:
            raise UnistrantError("Command failed with errors")


def main() -> None:
    try:
        _main()
    except KeyboardInterrupt:
        sys.exit(1)
    except UnistrantError as e:
        logger.error(str(e))
        sys.exit(1)



import enum
import xml.etree.ElementTree as ET
from collections.abc import Collection
from pathlib import Path

from unistrant.error import RecordError


class RecordType(enum.StrEnum):
    Compute = "Compute"
    Storage = "Storage"
    Cloud = "Cloud"
    SoftwareAccounting = "SoftwareAccounting"


class RecordTag(enum.StrEnum):
    JobUsageRecord = "{http://schema.ogf.org/urf/2003/09/urf}JobUsageRecord"
    StorageUsageRecord = "{http://eu-emi.eu/namespaces/2011/02/storagerecord}StorageUsageRecord"
    CloudComputeRecord = "{http://sams.snic.se/namespaces/2016/04/cloudrecords}CloudComputeRecord"
    SoftwareAccountingRecord = "{http://sams.snic.se/namespaces/2019/01/softwareaccountingrecords}SoftwareAccountingRecord"


class RecordCollectionTag(enum.StrEnum):
    UsageRecords = "{http://schema.ogf.org/urf/2003/09/urf}UsageRecords"
    StorageUsageRecords = "{http://eu-emi.eu/namespaces/2011/02/storagerecord}StorageUsageRecords"
    CloudRecords = "{http://sams.snic.se/namespaces/2016/04/cloudrecords}CloudRecords"
    SoftwareAccountingRecords = "{http://sams.snic.se/namespaces/2019/01/softwareaccountingrecords}SoftwareAccountingRecords"


class Record:
    def __init__(self, element: ET.Element):
        self.element = element

        try:
            self.tag = RecordTag(element.tag)
        except ValueError:
            raise RecordError(f"Unsupported tag {element.tag}")

    @property
    def record_type(self) -> RecordType:
        match self.tag:
            case RecordTag.JobUsageRecord:
                return RecordType.Compute
            case RecordTag.StorageUsageRecord:
                return RecordType.Storage
            case RecordTag.CloudComputeRecord:
                return RecordType.Cloud
            case RecordTag.SoftwareAccountingRecord:
                return RecordType.SoftwareAccounting


class RecordFile:
    def __init__(self, path: Path):
        with path.open("rb") as f:
            try:
                tree = ET.parse(f)
            except ET.ParseError as e:
                raise RecordError(f"Error parsing file {path.name}: {str(e)}")
        ET.indent(tree)
        self.path = path
        self.tree = tree

        self.tag: RecordTag | RecordCollectionTag
        tag = self.tree.getroot().tag
        try:
            self.tag = RecordTag(tag)
        except ValueError:
            try:
                self.tag = RecordCollectionTag(tag)
            except ValueError:
                raise RecordError(f"Unsupported tag {tag}")

    @property
    def name(self) -> str:
        return self.path.name

    @property
    def records(self) -> Collection[Record]:
        match self.tag:
            case RecordTag():
                return [Record(self.tree.getroot())]

            case RecordCollectionTag.UsageRecords:
                return [Record(element) for element in self.tree.findall(RecordTag.JobUsageRecord)]

            case RecordCollectionTag.StorageUsageRecords:
                return [Record(element) for element in self.tree.findall(RecordTag.StorageUsageRecord)]

            case RecordCollectionTag.CloudRecords:
                return [Record(element) for element in self.tree.findall(RecordTag.CloudComputeRecord)]

            case RecordCollectionTag.SoftwareAccountingRecords:
                return [Record(element) for element in self.tree.findall(RecordTag.SoftwareAccountingRecord)]


class RecordDocument:
    def __init__(self, record_type: RecordType, records: Collection[Record]):
        self.record_type = record_type
        self.records = records
        match record_type:
            case RecordType.Compute:
                tag = RecordCollectionTag.UsageRecords
            case RecordType.Storage:
                tag = RecordCollectionTag.StorageUsageRecords
            case RecordType.Cloud:
                tag = RecordCollectionTag.CloudRecords
            case RecordType.SoftwareAccounting:
                tag = RecordCollectionTag.SoftwareAccountingRecords
        tree = ET.Element(str(tag))
        for record in records:
            tree.append(record.element)
        ET.indent(tree)
        self.tree = tree

    @property
    def xml(self) -> str:
        return ET.tostring(self.tree, encoding="unicode")

    @property
    def bytes(self) -> bytes:
        return ET.tostring(self.tree)



import xml.etree.ElementTree as ET
from collections.abc import Mapping

from unistrant.http import HttpProtocol
from unistrant.record import RecordDocument, RecordType


def service_for_record(record_type: RecordType) -> str:
    match record_type:
        case RecordType.Compute:
            return "Registration"
        case RecordType.Storage:
            return "StorageRegistration"
        case RecordType.Cloud:
            return "CloudRegistration"
        case RecordType.SoftwareAccounting:
            return "SoftwareAccountingRegistration"


class SamsClient:
    def __init__(self, url: str, protocol: HttpProtocol) -> None:
        self.url = url
        self.protocol = protocol
        self._service_map: Mapping[str, str] = {}

    @property
    def service_map(self) -> Mapping[str, str]:
        if not self._service_map:
            self._service_map = self.fetch_service_map()
        return self._service_map

    def fetch_service_map(self) -> Mapping[str, str]:
        content = self.protocol.get(self.url)

        service_map = {}
        tree = ET.fromstring(content)
        for service in tree.findall("service"):
            if (element := service.find("name")) is not None:
                name = element.text
            else:
                continue
            if (element := service.find("href")) is not None:
                href = element.text
            else:
                continue
            if name and href:
                service_map[name] = href
        return service_map

    def upload_record_document(self, document: RecordDocument) -> None:
        service_name = service_for_record(document.record_type)
        self.protocol.post(self.service_map[service_name], document.bytes)



import unittest
from collections.abc import Mapping
from unittest.mock import MagicMock

from unistrant.sams import SamsClient


class SamsClientTestCase(unittest.TestCase):
    def test_get_service_map_1(self) -> None:
        content = b'<?xml version="1.0" encoding="UTF-8"?><services></services>'
        protocol = MagicMock()
        protocol.get = MagicMock(return_value=content)
        client = SamsClient("https://sams.example.org:6143/sgas", protocol)
        service_map = client.fetch_service_map()
        expected: Mapping[str, str] = {}
        self.assertDictEqual(service_map, expected)

    def test_get_service_map_2(self) -> None:
        content = (
            b'<?xml version="1.0" encoding="UTF-8"?>'
            b"<services>"
            b"<service>"
            b"<name>Registration</name>"
            b"<href>https://sams.example.org:6143/sgas/ur</href>"
            b"</service>"
            b"<service>"
            b"<name>StorageRegistration</name>"
            b"</service>"
            b"</services>"
        )
        protocol = MagicMock()
        protocol.get = MagicMock(return_value=content)
        client = SamsClient("https://sams.example.org:6143/sgas", protocol)
        service_map = client.fetch_service_map()
        expected = {
            "Registration": "https://sams.example.org:6143/sgas/ur",
        }
        self.assertDictEqual(service_map, expected)

    def test_get_service_map_3(self) -> None:
        content = (
            b'<?xml version="1.0" encoding="UTF-8"?>'
            b"<services>"
            b"<service>"
            b"<name>Registration</name>"
            b"<href>https://sams.example.org:6143/sgas/ur</href>"
            b"</service>"
            b"<service>"
            b"<href>https://sams.example.org:6143/sgas/sr</href>"
            b"</service>"
            b"</services>"
        )
        protocol = MagicMock()
        protocol.get = MagicMock(return_value=content)
        client = SamsClient("https://sams.example.org:6143/sgas", protocol)
        service_map = client.fetch_service_map()
        expected = {
            "Registration": "https://sams.example.org:6143/sgas/ur",
        }
        self.assertDictEqual(service_map, expected)

    def test_get_service_map_4(self) -> None:
        content = (
            b'<?xml version="1.0" encoding="UTF-8"?>'
            b"<services>"
            b"<service>"
            b"<name>Registration</name>"
            b"<href>https://sams.example.org:6143/sgas/ur</href>"
            b"</service>"
            b"<service>"
            b"<name>StorageRegistration</name>"
            b"<href>https://sams.example.org:6143/sgas/sr</href>"
            b"</service>"
            b"</services>"
        )
        protocol = MagicMock()
        protocol.get = MagicMock(return_value=content)
        client = SamsClient("https://sams.example.org:6143/sgas", protocol)
        service_map = client.fetch_service_map()
        expected = {
            "Registration": "https://sams.example.org:6143/sgas/ur",
            "StorageRegistration": "https://sams.example.org:6143/sgas/sr",
        }
        self.assertDictEqual(service_map, expected)
